package opentelemetry-client

  1. Overview
  2. Docs

Module Opentelemetry_clientSource

Sourcemodule Bounded_queue : sig ... end

Interface for a thread-safe, bounded queue.

Sourcemodule Common_ : sig ... end
Sourcemodule Consumer : sig ... end

Consumer that accepts items from a bounded queue and processes them.

Sourcemodule Emitter_combine : sig ... end

Combine multiple emitters into one.

Sourcemodule Emitter_limit_interval : sig ... end

Limit frequency at which the emitter emits.

Sourcemodule Emitter_sample : sig ... end
Sourcemodule Export_error : sig ... end

Error that can occur during export

Sourcemodule Exporter_combine : sig ... end

Combine multiple exporters into one

Sourcemodule Exporter_config : sig ... end

Constructing and managing the configuration common to many (most?) HTTP-based clients.

Sourcemodule Exporter_debug : sig ... end

Basic debug exporter, prints signals on stdout/stderr/...

Sourcemodule Exporter_queued : sig ... end

Build an exporter from a queue and a consumer.

Sourcemodule Exporter_stdout : sig ... end

A simple exporter that prints on stdout.

Sourcemodule Generic_consumer : sig ... end

A consumer: pulls signals from a queue, sends them somewhere else

Sourcemodule Generic_consumer_exporter : sig ... end

A consumer that just calls another exporter.

Sourcemodule Generic_http_consumer : sig ... end
Sourcemodule Generic_io : sig ... end

Generic IO monad.

Sourcemodule Generic_notifier : sig ... end

Generic notifier (used to signal when a bounded queue is empty)

Sourcemodule Http_config : sig ... end
Sourcemodule Resource_signal : sig ... end

Constructing and managing OTel signals at the resource (batch) level

Sourcemodule Sampler : sig ... end

Basic random sampling.

Sourcemodule Self_trace : sig ... end

Mini tracing module for OTEL itself.

Sourcemodule Util_net_backoff : sig ... end

Backoff behavior in case of errors

Sourcemodule Util_resources : sig ... end

Group signals into resource_xxx objects