package opentelemetry-client

  1. Overview
  2. Docs

Module Opentelemetry_client.Emitter_sampleSource

add_sampler sampler e is a new emitter that uses the sampler on each individual signal before passing them to e. This means only Sampler.proba_accept sampler of the signals will actually be emitted.

sample ~proba_accept e is add_sampler (Sampler.create ~proba_accept ()) e